
/***************************************************************************
 *
 * IPAT - Interferometer Threshold Curve Plotter
 *
 * Stephen R. Whiteley (stevew@srware.com)
 * 
 * Copyright (c) 2002 Whiteley Research Inc.  All Rights Reserved. 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************
 * $Id: ipat.cc,v 2.8 2009/01/11 21:07:16 stevew Exp $
 ***************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifdef WIN32_NATIVE
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <richedit.h>
#include <resource.h>
#else
#include <gtk/gtk.h>
#endif
#include "ipat.h"

// name of initialization file
// Format is:
//  #          (comments have '#' in first column)
//  2          (number of junctions, any text may follow number)
//  .8         (coupling factor, any text may follow number)
//  lpr -h     (printer command, or to file if space only)
//
#define IPATRC ".ipatrc"

//
// Global variables
//
const char *OUTNAME  = "ipat_plot.ps";   // default PostScript file
const char *DATANAME = ".ipat.data";     // data file for save/recall
const char *PRINTCMD = "lpr -h";         // default printer command
extern char   *help_text;

// Whiteley Research Logo bitmap
//
// XPM
static const char *tm30[] = {
    // width height ncolors chars_per_pixel
    "30 20 2 1",
    // colors
    "   c none",   
    ".  c blue",
    // pixels
    "                              ",
    "                              ",
    "   ..             ......      ",
    "  ...      .      ........    ",
    "  ....     ..    ..........   ",
    "   ...     ..    ...  .....   ",
    "   ....   ....   ...    ...   ",  
    "    ....  ....  ....    ...   ",
    "    ....  ..... ...........   ",
    "     .......... ...........   ",
    "     .....................    ",
    "      ...... ...........      ",
    "      .....  .....  ....      ",
    "       ....   ....   ....     ",
    "       ....   ....   .....    ",
    "        ..     ...    .....   ",
    "        ..      .      ....   ",
    "         .      .       ..    ",
    "                              ",
    "                              "};


// menu dispatch codes
enum { CODE_NONE, CODE_DLOG_SAVE, CODE_DLOG_DEL, CODE_DLOG_LOAD, CODE_PRINT,
    CODE_THREEJ, CODE_PHASE, CODE_X, CODE_Y, CODE_AXES, CODE_SECOND,
    CODE_TERT, CODE_FAIL, CODE_MARK };

// dialog popup configuration code
enum { DLOG_NONE, DLOG_SAVE, DLOG_DEL, DLOG_LOAD };

// id code from the spin buttons
enum { VAL_NONE, VAL_L1, VAL_L2, VAL_L3, VAL_L4, VAL_FRAC, VAL_IC1, VAL_IC2,
VAL_IC3, VAL_COUP };

// The application class
panel app;

// The parameters
values vals;

extern clr Colors[];
static char *get_data_file(const char*);
static void errmsg(const char*, int);


int
main(int argc, char **argv)
{
    static char buf[84];
    signal(SIGFPE,SIG_IGN);

    FILE *fp = fopen(IPATRC, "r");
    if (!fp) {
        char *s = getenv("HOME");
        if (s) {
            char *t = s + strlen(s) - 1;
            if (*t == '/'
#ifdef WIN32
                    || *t == '\\'
#endif
                    )
                sprintf(buf, "%s%s",  s, IPATRC);
            else
                sprintf(buf, "%s/%s", s, IPATRC);
            fp = fopen(buf, "r");
        }
    }
    if (fp) {
        char *s;
        while ((s = fgets(buf, 84, fp)) != NULL && *s == '#');
        if (s) {
            int i = atoi(s);
            if (i != 2 && i != 3)
                i = 2;
            app.flags.threej = i - 2;
            if (app.flags.threej)
                vals.i2 = 2.0;
        }
        while ((s = fgets(buf, 84, fp)) != NULL && *s == '#');
        if (s) {
            double d = atof(s);
            if (d > 0.0)
                app.Coupling = d;
        }
        while ((s = fgets(buf, 84, fp)) != NULL && *s == '#');
        if (s) {
            *(s + strlen(s) - 1) = '\0';  // zap newline
            while (isspace(*s))
                s++;
#ifdef WIN32
            // Ignore this in Windows.
#else
            if (*s)
                app.cmdtext = strdup(s);
            else {
                app.cmdtext = strdup(PRINTCMD);
                app.tofilename = strdup(OUTNAME);
                app.tofile = true;
            }
#endif
        }
        fclose(fp);
    }

    int i = 1;
    while (argc-- > 1) {
        fp = fopen(argv[i++],"r");
        if (fp)
            app.extpts(fp);
    }

    // This contains the main loop
    if (!app.init())
        return (1);
    return (0);
}


#ifdef WIN32_NATIVE

#define ENTHT 36
#define WINHT 620
#define WINWD 510
typedef ATOM (WINAPI* REGISTERCLASSEXPROC)(const LPWNDCLASSEX);

static ATOM msw_RegisterClass(const LPWNDCLASSEX);
static bool get_status(int);
static void set_status(int, bool);
static void resize();
static LRESULT CALLBACK mainFrameWndProc(HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK drawingWndProc(HWND, UINT, WPARAM, LPARAM);
static HFONT msw_GetFixedFont();
static bool xpm2dib(const char**, BITMAPINFOHEADER**, unsigned char**);
static HBITMAP msw_DIBfromXPM(HWND, const char**);
static void create_entry(HWND, int, int, spinbtn*, const char*);

#else

#define IFINIT(i, a, b, c, d, e) { \
    menu_items[i].path = (char*)a; \
    menu_items[i].accelerator = (char*)b; \
    menu_items[i].callback = (GtkItemFactoryCallback)c; \
    menu_items[i].callback_action = d; \
    menu_items[i].item_type = (char*)e; \
    i++; }

static void cancel_proc(GtkWidget*, void*);
static void help_proc(GtkWidget*, void*);
static int da1_expose_hdlr(GtkWidget*, GdkEvent*, void*);
static int da2_expose_hdlr(GtkWidget*, GdkEvent*, void*);
static int motion_hdlr(GtkWidget*, GdkEvent*, void*);
static int leave_hdlr(GtkWidget*, GdkEvent*, void*);
static int button_dn_hdlr(GtkWidget*, GdkEvent*, void*);
static int key_dn_hdlr(GtkWidget*, GdkEvent*, void*);
static void wr_btn_hdlr(GtkWidget*, void*);
static void menu_hdlr(GtkWidget*, void*, unsigned int);
static int redraw_idle_proc(void*);
static void queue_redraw();
static void value_changed(GtkWidget*, void*);
static int button_press(GtkWidget*, GdkEvent*);
static void recall_menu_hdlr(GtkWidget*, void*);

#endif


// Main initialization for the user interface, return false if error
//
bool
panel::init()
{
#ifdef WIN32_NATIVE

    HINSTANCE instance = GetModuleHandle(0);

    // Register all of the application's window classes.
    //
    // The application's main frame window
    //
    WNDCLASSEX wcex;
    wcex.cbSize        = sizeof(WNDCLASSEX);
    wcex.style         = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc   = mainFrameWndProc;
    wcex.cbClsExtra    = 0;
    wcex.cbWndExtra    = 0;
    wcex.hInstance     = instance;
    wcex.hIcon         = LoadIcon(instance, MAKEINTRESOURCE(IDR_MAINFRAME));
    wcex.hCursor       = LoadCursor(0, IDC_ARROW);
    wcex.hbrBackground = GetStockBrush(LTGRAY_BRUSH);
    wcex.lpszMenuName  = MAKEINTRESOURCE(IDR_MAINFRAME);
    wcex.lpszClassName = MAKEINTRESOURCE(IDR_MAINFRAME);
    wcex.hIconSm       = (HICON)LoadImage(instance,
        MAKEINTRESOURCE(IDR_MAINFRAME), IMAGE_ICON,
        GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON),
        LR_SHARED);

    if (!msw_RegisterClass(&wcex))
        return (false);

    // The drawing window class
    //
    wcex.style         = CS_HREDRAW | CS_VREDRAW | CS_CLASSDC;
    wcex.lpfnWndProc   = drawingWndProc;
    wcex.cbClsExtra    = 0;
    wcex.cbWndExtra    = 0;
    wcex.hInstance     = instance;
    wcex.hIcon         = 0;
    wcex.hCursor       = LoadCursor(0, IDC_ARROW);
    wcex.hbrBackground = GetStockBrush(BLACK_BRUSH);
    wcex.lpszMenuName  = 0;
    wcex.lpszClassName = MAKEINTRESOURCE(IDR_DRAWINGWIN);
    wcex.hIconSm       = 0;

    if (!msw_RegisterClass(&wcex))
        return (false);

    // Create the main frame window
    const char *title = "IPAT";
    shell = CreateWindowEx(
        0,                  // Extended window styles
        MAKEINTRESOURCE(IDR_MAINFRAME), // registered class name
        title,              // Address of window name
        WS_OVERLAPPEDWINDOW,// Window style
        CW_USEDEFAULT,      // Horizontal position of window
        0,                  // Vertical position of window
        WINWD,              // Window width
        WINHT,              // Window height
        0,                  // Handle of parent or owner window
        0,                  // Handle of menu for this window
        instance,           // Handle of application instance
        0);                 // Address of window-creation data

    if (!shell)
       return (false);

    // Create the main drawing child window
    scr.da = CreateWindowEx( 
       WS_EX_NOPARENTNOTIFY | WS_EX_ACCEPTFILES,
       MAKEINTRESOURCE(IDR_DRAWINGWIN),  // Class name
       0,                          // Child caption
       WS_CHILDWINDOW | WS_BORDER | WS_VISIBLE, // window style
       60,                         // origin
       0,                          // ...
       440,                        // size
       9*ENTHT,                    // ...
       shell,                      // owner
       (HMENU)IDR_DRAWINGWIN,      // no menu
       instance,                   // Instance handle
       0);                         // user-defined parameters

    if (scr.da == 0)
        return (false);

    diag.da = CreateWindowEx( 
       WS_EX_NOPARENTNOTIFY | WS_EX_ACCEPTFILES,
       MAKEINTRESOURCE(IDR_DRAWINGWIN),  // Class name
       0,                          // Child caption
       WS_CHILDWINDOW | WS_BORDER | WS_VISIBLE, // window style
       0,                          // origin
       9*ENTHT,                    // ...
       WINWD,                      // size
       WINHT - 9*ENTHT,            // ...
       shell,                      // owner
       (HMENU)IDR_DRAWINGWIN,      // no menu
       instance,                   // Instance handle
       0);                         // user-defined parameters

    if (diag.da == 0)
        return (false);

    InitCommonControls();
    if (!LoadLibrary("RICHED32.DLL")) {
        errmsg("Error: No access to \"RICHEDIT.DLL\".", 1);
        return (false);
    }
    int ypos = 0;
    int res = 200;
    create_entry(shell, res, ypos, &entries[0], "L1");
    entries[0].init(vals.la, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[1], "L2");
    entries[1].init(vals.lb, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[2], "L3");
    entries[2].init(vals.lc, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[3], "L4");
    entries[3].init(vals.ld, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[4], "Frac");
    entries[4].init(vals.lfrac, 0.01, 0.01, 1.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[5], "I1");
    entries[5].init(vals.ix, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[6], "I2");
    entries[6].init(vals.i2, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[7], "I3");
    entries[7].init(vals.iy, 0.01, 0.01, 100.0, 2);
    res += 3;
    ypos += ENTHT;
    create_entry(shell, res, ypos, &entries[8], "Coupling");
    entries[8].init(app.Coupling, 0.01, 0.01, 100.0, 2);

    // Insert the WR button
    HBITMAP bitmap = msw_DIBfromXPM(shell, tm30);
    HMENU menu = GetMenu(shell);
    InsertMenu(menu, 0, MF_BYPOSITION | MF_BITMAP, ID_WR_BTN, (LPCTSTR)bitmap);

    // Initialize the Recall menu
    init_recall_menu();

    //
    // create GCs and fonts
    //
    scr.gc = GetDC(scr.da);
    scr.font = msw_GetFixedFont();

    //
    // initialize colors
    //
    for (int i = 1; i < 16; i++) {
        int red = Colors[i].red;
        int green = Colors[i].green;
        int blue = Colors[i].blue;
        scr.define_color(i, red, green, blue);
    }

    SetBkColor(scr.gc, RGB(0, 0, 0));
    SetTextColor(scr.gc, RGB(255, 255, 255));
    SetBkMode(scr.gc, TRANSPARENT);
    SelectFont(scr.gc, scr.font);
    SetTextAlign(scr.gc, TA_LEFT | TA_BOTTOM | TA_NOUPDATECP);

    //
    // initialize plot area
    //
    SIZE size;
    GetTextExtentPoint32(scr.gc, "x", 1, &size);
    int fheight = size.cy;
    int fwidth  = size.cx;
    scr.W   = fwidth;
    scr.H   = fheight;
    scr.W2  = fwidth/2;
    scr.H2  = fheight/2;
    scr.H3  = scr.H + scr.H2;
    scr.XL  = 4*fwidth;
    scr.YL  = 3*fheight;

    RECT r;
    GetClientRect(scr.da, &r);
    scr.XS = r.right - r.left;
    scr.YS = r.bottom - r.top;

    scr.X   = scr.XS - scr.XL - 2*fwidth;
    scr.Y   = scr.YS - scr.YL - 3*fheight;
    scr.XC  = scr.XL + scr.X/2;
    scr.YC  = scr.YL + scr.Y/2;
    scr.XU  = scr.XL + scr.X;
    scr.YU  = scr.YL + scr.Y;

    //
    // initialize diagram area
    //
    diag.gc = scr.gc;
    diag.font = scr.font;

    diag.W   = fwidth;
    diag.H   = fheight;
    diag.W2  = fwidth/2;
    diag.H2  = fheight/2;
    diag.H3  = diag.H + diag.H2;
    diag.XL  = 4*fwidth;
    diag.YL  = 3*fheight;

    GetClientRect(diag.da, &r);
    diag.XS = r.right - r.left;
    diag.YS = r.bottom - r.top;

    diag.X   = diag.XS - diag.XL - 2*fwidth;
    diag.Y   = diag.YS - diag.YL - 3*fheight;
    diag.XC  = diag.XL + diag.X/2;
    diag.YC  = diag.YL + diag.Y/2;
    diag.XU  = diag.XL + diag.X;
    diag.YU  = diag.YL + diag.Y;

    char buf[128];
    sprintf(buf, "IPAT %s - Whiteley Research Inc.", VERSION);
    SetWindowText(shell, buf);
    ShowWindow(shell, SW_SHOW);

    MSG msg;
    while (GetMessage(&msg, 0, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

#else
    gtk_init(0, 0);

    // do this before pixmab creation below
    scr.cmap = gdk_colormap_get_system();

    shell = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    char buf[128];
    sprintf(buf, "IPAT %s - Whiteley Research Inc.", VERSION);
    gtk_window_set_title(GTK_WINDOW(shell), buf);
    gtk_signal_connect(GTK_OBJECT(shell), "destroy",
        GTK_SIGNAL_FUNC(cancel_proc), 0);
    gtk_signal_connect_after(GTK_OBJECT(shell), "key_press_event",
        GTK_SIGNAL_FUNC(key_dn_hdlr), &scr);


    GtkWidget *form = gtk_table_new(2, 3, false);
    gtk_widget_show(form);
    gtk_container_add(GTK_CONTAINER(shell), form);

    //
    // WR button
    //
    GtkWidget *pixbtn = gtk_button_new();
    gtk_widget_show(pixbtn);
    gtk_signal_connect(GTK_OBJECT(pixbtn), "clicked",
        GTK_SIGNAL_FUNC(wr_btn_hdlr), 0);
    GtkStyle *style = gtk_widget_get_style(pixbtn);
    GdkPixmap *pmask;   
    GdkPixmap *pixmap =
        gdk_pixmap_colormap_create_from_xpm_d(0, scr.cmap,
            &pmask, &style->bg[GTK_STATE_NORMAL], (gchar **)tm30);
    GtkWidget *pixwidg = gtk_pixmap_new(pixmap, pmask);
    gtk_widget_show(pixwidg);
    gtk_container_add(GTK_CONTAINER(pixbtn), pixwidg);

    gtk_table_attach(GTK_TABLE(form), pixbtn, 0, 1, 0, 1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 0, 0);

    //
    // menu bar
    //
    GtkItemFactoryEntry menu_items[20];
    int nitems = 0;

    IFINIT(nitems, "/_File", 0, 0, 0, "<Branch>")
    IFINIT(nitems, "/File/_Save", 0, menu_hdlr, CODE_DLOG_SAVE, 0);
    IFINIT(nitems, "/File/_Delete", 0, menu_hdlr, CODE_DLOG_DEL, 0);
    IFINIT(nitems, "/File/_Recall", 0, 0, 0, 0);
    IFINIT(nitems, "/File/_Load Points", 0, menu_hdlr, CODE_DLOG_LOAD, 0);
    IFINIT(nitems, "/File/_Print", "<alt>P", menu_hdlr, CODE_PRINT, 0);
    IFINIT(nitems, "/File/sep1", 0, 0, 0, "<Separator>");
    IFINIT(nitems, "/File/_Quit", "<control>Q", cancel_proc, 0, 0);
    IFINIT(nitems, "/_Options", 0, 0, 0, "<Branch>");
    IFINIT(nitems, "/_Options/Three _Junctions", 0, menu_hdlr, CODE_THREEJ,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/_Phase Plot", 0, menu_hdlr, CODE_PHASE,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Show Negative _X", 0, menu_hdlr, CODE_X,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Show Negative _Y", 0, menu_hdlr, CODE_Y,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Suppress _Axes", 0, menu_hdlr, CODE_AXES,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Suppress _Secondary", 0, menu_hdlr, CODE_SECOND,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Suppress _Tertiary", 0, menu_hdlr, CODE_TERT,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Plot _Fail Points", 0, menu_hdlr, CODE_FAIL,
        "<CheckItem>");
    IFINIT(nitems, "/_Options/Show _Marker", 0, menu_hdlr, CODE_MARK, 0);
    IFINIT(nitems, "/_Help", 0, 0, 0, "<LastBranch>");
    IFINIT(nitems, "/Help/_Help", "<alt>H", help_proc, 0, 0);

    GtkAccelGroup *accel_group = gtk_accel_group_new();
    item_factory =
        gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<ipat>", accel_group);
    for (int i = 0; i < nitems; i++)
        gtk_item_factory_create_item(item_factory, menu_items + i, 0, 2);

#if GTK_CHECK_VERSION(2,0,0)
    gtk_window_add_accel_group(GTK_WINDOW(shell), accel_group);
#else
    gtk_accel_group_attach(accel_group, GTK_OBJECT(shell));
#endif

    menubar = gtk_item_factory_get_widget(item_factory, "<ipat>");
    gtk_widget_show(menubar);

    gtk_table_attach(GTK_TABLE(form), menubar, 1, 2, 0, 1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 0, 0);

    // Initialize the Recall menu
    init_recall_menu();

    //
    // plot area
    //
    scr.da = gtk_drawing_area_new();
    gtk_widget_show(scr.da);
    gtk_drawing_area_size(GTK_DRAWING_AREA(scr.da), 500, 250);
    gtk_signal_connect(GTK_OBJECT(scr.da), "expose_event",
        GTK_SIGNAL_FUNC(da1_expose_hdlr), &scr);
    gtk_widget_add_events(scr.da, GDK_BUTTON_PRESS_MASK);
    gtk_widget_add_events(scr.da, GDK_POINTER_MOTION_MASK);
    gtk_signal_connect(GTK_OBJECT(scr.da), "motion_notify_event",
        GTK_SIGNAL_FUNC(motion_hdlr), &scr);
    gtk_widget_add_events(scr.da, GDK_LEAVE_NOTIFY_MASK);
    gtk_signal_connect(GTK_OBJECT(scr.da), "leave_notify_event",
        GTK_SIGNAL_FUNC(leave_hdlr), &scr);
    gtk_signal_connect_after(GTK_OBJECT(scr.da), "button_press_event", 
        GTK_SIGNAL_FUNC(button_dn_hdlr), &scr);
    GtkWidget *frame = gtk_frame_new(0);
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), scr.da);
    
    gtk_table_attach(GTK_TABLE(form), frame, 1, 2, 1, 2,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 0, 0);

    //
    // input widgets
    //
    GtkWidget *vbox = gtk_vbox_new(false, 2);
    gtk_widget_show(vbox);

    GtkObject *adj = gtk_adjustment_new(vals.la, .01, 100.0, .01, 1.0, 0);
    GtkWidget *entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("L1");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_L1);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_l1 = entry;

    adj = gtk_adjustment_new(vals.lb, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("L2");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_L2);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_l2 = entry;

    adj = gtk_adjustment_new(vals.lc, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("L3");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_L3);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_l3 = entry;

    adj = gtk_adjustment_new(vals.ld, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("L4");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_L4);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_l4 = entry;

    adj = gtk_adjustment_new(vals.lfrac, 0.0, 1.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("Frac");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_FRAC);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_frac = entry;

    adj = gtk_adjustment_new(vals.ix, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("Ic1");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_IC1);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_i1 = entry;

    adj = gtk_adjustment_new(vals.i2, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("Ic2");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_IC2);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_i2 = entry;

    adj = gtk_adjustment_new(vals.iy, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("Ic3");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_IC3);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_i3 = entry;

    adj = gtk_adjustment_new(app.Coupling, .01, 100.0, .01, 1.0, 0);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 2);
    gtk_widget_show(entry);
    frame = gtk_frame_new("Coupling");
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), entry);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(entry), true);
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
        GTK_SIGNAL_FUNC(value_changed), (void*)VAL_COUP);
    gtk_box_pack_start(GTK_BOX(vbox), frame, false, false, 2);
    entry_coup = entry;

    gtk_table_attach(GTK_TABLE(form), vbox, 0, 1, 1, 2,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 0, 0);

    //
    // diagram area
    //
    diag.da = gtk_drawing_area_new();
    gtk_widget_show(diag.da);
    gtk_drawing_area_size(GTK_DRAWING_AREA(diag.da), 600, 250);
    gtk_signal_connect(GTK_OBJECT(diag.da), "expose_event",
        GTK_SIGNAL_FUNC(da2_expose_hdlr), 0);
    frame = gtk_frame_new(0);
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), diag.da);

    gtk_table_attach(GTK_TABLE(form), frame, 0, 2, 2, 3,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 0, 0);

    gtk_widget_show(shell);

    //
    // create GCs and fonts
    //
    scr.gc = gdk_gc_new(scr.da->window);
    GdkGCValues gcvalues;
    gcvalues.function = GDK_XOR;
    scr.xorgc = gdk_gc_new_with_values(scr.da->window, &gcvalues,
        (GdkGCValuesMask)GDK_GC_FUNCTION);
    scr.font = gdk_font_load("fixed");

    //
    // initialize colors
    //
    for (int i = 0; i < 16; i++) {
        int red = Colors[i].red;
        int green = Colors[i].green;
        int blue = Colors[i].blue;
        scr.define_color(i, red, green, blue);
    }

    GdkColor clr;
    clr.pixel = Colors[0].pixel;
    gdk_window_set_background(scr.da->window, &clr);
    gdk_window_set_background(diag.da->window, &clr);
    gdk_gc_set_background(scr.gc, &clr);
    gdk_gc_set_background(scr.xorgc, &clr);
    clr.pixel = Colors[0].pixel ^ Colors[1].pixel;
    gdk_gc_set_foreground(scr.xorgc, &clr);

    //
    // initialize plot area
    //
    int fheight = scr.font->ascent + scr.font->descent;
    int fwidth  = gdk_string_width(scr.font, "x");
    scr.W   = fwidth;
    scr.H   = fheight;
    scr.W2  = fwidth/2;
    scr.H2  = fheight/2;
    scr.H3  = scr.H + scr.H2;
    scr.XL  = 4*fwidth;
    scr.YL  = 3*fheight;
    gdk_window_get_size(scr.da->window, &scr.XS, &scr.YS);
    scr.X   = scr.XS - scr.XL - 2*fwidth;
    scr.Y   = scr.YS - scr.YL - 3*fheight;
    scr.XC  = scr.XL + scr.X/2;
    scr.YC  = scr.YL + scr.Y/2;
    scr.XU  = scr.XL + scr.X;
    scr.YU  = scr.YL + scr.Y;

    //
    // initialize diagram area
    //
    diag.gc = scr.gc;
    diag.xorgc = scr.xorgc;
    diag.font = scr.font;
    diag.cmap = scr.cmap;

    diag.W   = fwidth;
    diag.H   = fheight;
    diag.W2  = fwidth/2;
    diag.H2  = fheight/2;
    diag.H3  = diag.H + diag.H2;
    diag.XL  = 4*fwidth;
    diag.YL  = 3*fheight;
    gdk_window_get_size(diag.da->window, &diag.XS, &diag.YS);
    diag.X   = diag.XS - diag.XL - 2*fwidth;
    diag.Y   = diag.YS - diag.YL - 3*fheight;
    diag.XC  = diag.XL + diag.X/2;
    diag.YC  = diag.YL + diag.Y/2;
    diag.XU  = diag.XL + diag.X;
    diag.YU  = diag.YL + diag.Y;

    gtk_main();
#endif

    return (true);
}


// Initialize the Recall menu of saved data sets
//
void
panel::init_recall_menu()
{
#ifdef WIN32_NATIVE
    MENUITEMINFO mii;
    mii.cbSize = sizeof(MENUITEMINFO);
    mii.fMask = MIIM_SUBMENU;
    HMENU menu = GetMenu(shell);
    if (!GetMenuItemInfo(menu, 1, true, &mii) || !mii.hSubMenu)
        return;
    HMENU hmenu = mii.hSubMenu;
    hmenu = GetSubMenu(hmenu, 2);
    if (!hmenu)
        return;

    int cnt = GetMenuItemCount(hmenu);
    while (cnt--)
        RemoveMenu(hmenu, 0, MF_BYPOSITION);

    char *datafile = get_data_file(DATANAME);
    FILE *fp = fopen(datafile, "r");
    if (fp) {
        char buf[256];
        while (fgets(buf, 256, fp) != 0) {
            char *s = buf;
            while (isspace(*s))
                s++;
            if (!*s || *s == '#')
                continue;
            char *tok = strdup(buf);
            s = tok;
            while (*s && !isspace(*s))
                s++;
            *s = 0;
            AppendMenu(hmenu, MF_STRING, ID_RECALL + GetMenuItemCount(hmenu),
                tok);
            delete [] tok;
        }
        fclose(fp);
    }
    delete [] datafile;

#else
    GtkWidget *root =
        gtk_item_factory_get_widget(item_factory, "/File/Recall");
    if (!root)
        return;
    GtkWidget *menu = gtk_menu_new();

    char *datafile = get_data_file(DATANAME);
    FILE *fp = fopen(datafile, "r");
    if (fp) {
        char buf[256];
        while (fgets(buf, 256, fp) != 0) {
            char *s = buf;
            while (isspace(*s))
                s++;
            if (!*s || *s == '#')
                continue;
            char *tok = strdup(buf);
            s = tok;
            while (*s && !isspace(*s))
                s++;
            *s = 0;

            GtkWidget *menu_item = gtk_menu_item_new_with_label(tok);
            gtk_object_set_data(GTK_OBJECT(menu_item), "name", strdup(tok));
            gtk_menu_append(GTK_MENU(menu), menu_item);
            gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
                GTK_SIGNAL_FUNC(recall_menu_hdlr), 0);
            gtk_widget_show(menu_item);
            delete [] tok;
        }
        fclose(fp);
    }
    delete [] datafile;
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(root), menu);
    gtk_signal_connect_object(GTK_OBJECT(root), "event",
        GTK_SIGNAL_FUNC(button_press), GTK_OBJECT(menu));
#endif
}


#ifdef WIN32_NATIVE

// Register the window class using RegisterClassEx if it is available.
// If not, registers the class using RegisterClass.
//
static ATOM
msw_RegisterClass(const LPWNDCLASSEX lpwcex)
{
    // Get the module handle of the 32-bit USER DLL
    HANDLE hModule = GetModuleHandle (TEXT("USER32"));
    if (hModule != 0) {

        // If we're running on a Win32 version supporting RegisterClassEx
        //  get the address of the function so we can call it

        REGISTERCLASSEXPROC proc =
            (REGISTERCLASSEXPROC)GetProcAddress((HINSTANCE)hModule, 
            "RegisterClassExA");

        if (proc != 0)
            // RegisterClassEx exists...
            // return RegisterClassEx(&wcex);
            return (*proc)(lpwcex);
    }

    // Convert the WNDCLASSEX structure to a WNDCLASS structure
    WNDCLASS wc;
    wc.style         = lpwcex->style;
    wc.lpfnWndProc   = lpwcex->lpfnWndProc;
    wc.cbClsExtra    = lpwcex->cbClsExtra;
    wc.cbWndExtra    = lpwcex->cbWndExtra;
    wc.hInstance     = lpwcex->hInstance;
    wc.hIcon         = lpwcex->hIcon;
    wc.hCursor       = lpwcex->hCursor;
    wc.hbrBackground = lpwcex->hbrBackground;
    wc.lpszMenuName  = lpwcex->lpszMenuName;
    wc.lpszClassName = lpwcex->lpszClassName;

    return (RegisterClass(&wc));
}


static bool
get_status(int id)
{
    HMENU menu = GetMenu(app.shell);
    MENUITEMINFO mii;
    mii.cbSize = sizeof(mii);
    mii.fMask = MIIM_STATE | MIIM_DATA;
    if (!GetMenuItemInfo(menu, id, false, &mii))
        return (true);
    return ((mii.fState & MFS_CHECKED) ? true : false);
}


static void
set_status(int id, bool state)
{
    HMENU menu = GetMenu(app.shell);
    if (menu)
        CheckMenuItem(menu, id, state ? MF_CHECKED : MF_UNCHECKED);
}


static void
resize()
{
    RECT r;
    GetClientRect(app.shell, &r);
    if (r.bottom - r.top < 9*ENTHT + 20) {
        RECT r1;
        GetWindowRect(app.shell, &r1);
        MoveWindow(app.shell, r1.left, r1.top, r1.right - r1.left,
            9*ENTHT + 20, false);
        r.bottom = 9*ENTHT + 20;
    }
    if (r.right - r.left < 100) {
        RECT r1;
        GetWindowRect(app.shell, &r1);
        MoveWindow(app.shell, r1.left, r1.top, 100, r1.bottom - r1.top, false);
        r.right = 100;
    }

    MoveWindow(app.scr.da, 60, 0, r.right - r.left - 60, 9*ENTHT, false);
    MoveWindow(app.diag.da, 0, 9*ENTHT, r.right - r.left,
        r.bottom - r.top - 9*ENTHT, false);

    GetClientRect(app.scr.da, &r);
    app.scr.XS = r.right - r.left;
    app.scr.YS = r.bottom - r.top;

    GetClientRect(app.diag.da, &r);
    app.diag.XS = r.right - r.left;
    app.diag.YS = r.bottom - r.top;
}


static LRESULT CALLBACK
mainFrameWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    NM_UPDOWN *udn;
    switch (message) {

    case WM_SIZE:
        resize();
        return (0);

    case WM_KEYDOWN:
        {
            screen *sc = &app.scr;
            if (sc->draw_ghost) {
                sc->refmark();
                if (!sc->first_ghost)
                    sc->marker(sc->last_x, sc->last_y);
                sc->draw_ghost = false;
                app.flags.rel_x = -1;
                app.flags.rel_y = -1;
                InvalidateRect(app.scr.da, 0, false);
                return (0);
            }
        }
        break;

    case WM_MOUSEMOVE:
        // these only come when the pointer is in the area outside of the
        // entry edit control but not in the drawing windows
        for (int i = 0; i < 9; i++) {
            char buf[64];
            Edit_GetText(app.entries[i].entry, buf, 64);
            double d = atof(buf);
            if (d < app.entries[i].minval)
                d = app.entries[i].minval;
            else if (d > app.entries[i].maxval)
                d = app.entries[i].maxval;
            sprintf(buf, "%.*f", app.entries[i].numdgt, d);
            Edit_SetText(app.entries[i].entry, buf);
            char buf1[64];
            sprintf(buf1, "%.*f", app.entries[i].numdgt, app.entries[i].value);
            if (strcmp(buf, buf1)) {
                app.entries[i].value = d;
                switch (i) {
                case 0:
                    vals.la = app.entries[i].value; break;
                case 1:
                    vals.lb = app.entries[i].value; break;
                case 2:
                    vals.lc = app.entries[i].value; break;
                case 3:
                    vals.ld = app.entries[i].value; break;
                case 4:
                    vals.lfrac = app.entries[i].value; break;
                case 5:
                    vals.ix = app.entries[i].value; break;
                case 6:
                    vals.i2 = app.entries[i].value; break;
                case 7:
                    vals.iy = app.entries[i].value; break;
                case 8:
                    app.Coupling = app.entries[i].value;
                    InvalidateRect(app.diag.da, 0, false);
                    break;
                }
                InvalidateRect(app.scr.da, 0, false);
            }
        }
        return (0);

    case WM_NOTIFY:
        udn = (NM_UPDOWN*)lParam;
        if (udn->hdr.code == (unsigned int)UDN_DELTAPOS) {
            for (int i = 0; i < 9; i++) {
                if (udn->hdr.hwndFrom == app.entries[i].ud) {
                    if (udn->iDelta > 0)
                        app.entries[i].setval(false);
                    else
                        app.entries[i].setval(true);
                    char buf[64];
                    Edit_GetText(app.entries[i].entry, buf, 64);
                    double d = atof(buf);
                    switch (i) {
                    case 0:
                        vals.la = d; break;
                    case 1:
                        vals.lb = d; break;
                    case 2:
                        vals.lc = d; break;
                    case 3:
                        vals.ld = d; break;
                    case 4:
                        vals.lfrac = d; break;
                    case 5:
                        vals.ix = d; break;
                    case 6:
                        vals.i2 = d; break;
                    case 7:
                        vals.iy = d; break;
                    case 8:
                        app.Coupling = d;
                        InvalidateRect(app.diag.da, 0, false);
                        break;
                    }
                    InvalidateRect(app.scr.da, 0, false);
                }
            }
            return (0);
        }
        break;

    case WM_COMMAND:
        if (LOWORD(wParam) == ID_WR_BTN) {
            static int num = 2;
            num++;
            if (num > 4)
                num = 1;
            app.Scale = 1 << (num - 1);
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_File_Save) {
            app.dialog_popup(DLOG_SAVE);
            return (0);
        }
        if (LOWORD(wParam) == ID_File_Delete) {
            app.dialog_popup(DLOG_DEL);
            return (0);
        }
        if (LOWORD(wParam) == ID_File_LoadPoints) {
            app.dialog_popup(DLOG_LOAD);
            return (0);
        }
        if (LOWORD(wParam) == ID_File_Print) {
            app.print_popup();
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_ThreeJunctions) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.threej = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.threej = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            InvalidateRect(app.diag.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_PhasePlot) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.z = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.z = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_ShowNegativeX) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.x = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.x = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_ShowNegativeY) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.y = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.y = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_SuppressAxes) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.s = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.s = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_SuppressSecondary) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.n = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.n = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_SuppressTertiary) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.r = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.r = true;
            }
            InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_PlotFailPoints) {
            if (get_status(LOWORD(wParam))) {
                set_status(LOWORD(wParam), false);
                app.flags.f = false;
            }
            else {
                set_status(LOWORD(wParam), true);
                app.flags.f = true;
            }
            if (app.pts)
                InvalidateRect(app.scr.da, 0, false);
            return (0);
        }
        if (LOWORD(wParam) == ID_Options_ShowMarker) {
            if (!app.scr.draw_ghost)
                app.scr.xy_pick();
            return (0);
        }
        if (LOWORD(wParam) == ID_Help_Help) {
            app.text_display_popup(help_text);
            return (0);
        }
        if (LOWORD(wParam) >= ID_RECALL && LOWORD(wParam) < ID_RECALL + 100) {
            char buf[256];
            MENUITEMINFO mii;
            mii.cbSize = sizeof(MENUITEMINFO);
            mii.fMask = MIIM_TYPE;
            mii.dwTypeData = buf;
            mii.cch = 256;
            HMENU menu = GetMenu(app.shell);
            if (GetMenuItemInfo(menu, LOWORD(wParam), false, &mii))
                app.recall(buf);
            return (0);
        }
        if (LOWORD(wParam) != ID_File_Quit)
            break;
        // fall thru

    case WM_CLOSE:
        exit (0);
    }
    return (DefWindowProc(hwnd, message, wParam, lParam));
}


static LRESULT CALLBACK
drawingWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {

    case WM_PAINT:
        if (hwnd == app.scr.da) {
            PAINTSTRUCT ps;
            BeginPaint(app.scr.da, &ps);
            EndPaint(app.scr.da, &ps);
            app.scr.displ();
            return (0);
        }
        else if (hwnd == app.diag.da) {
            PAINTSTRUCT ps;
            BeginPaint(app.diag.da, &ps);
            EndPaint(app.diag.da, &ps);
            app.diag.diagram();
            return (0);
        }
        break;

    case WM_MOUSEMOVE:
        if (hwnd == app.scr.da) {
            screen *sc = &app.scr;
            if (sc->draw_ghost) {
                if (!sc->first_ghost)
                    sc->marker(sc->last_x, sc->last_y);
                sc->last_x = LOWORD(lParam);
                sc->last_y = HIWORD(lParam);
                sc->marker(sc->last_x, sc->last_y);
                sc->first_ghost = false;
            }
        }
        break;

    case WM_LBUTTONDOWN:
        if (hwnd == app.scr.da) {
            screen *sc = &app.scr;
            if (sc->draw_ghost) {
                if (LOWORD(lParam) >= sc->XL && LOWORD(lParam) <= sc->XU &&
                        HIWORD(lParam) >= sc->YL && HIWORD(lParam) <= sc->YU) {
                    if (!sc->first_ghost) {
                        sc->marker(sc->last_x, sc->last_y);
                        sc->first_ghost = true;
                    }
                    sc->refmark();
                    app.flags.rel_x = (LOWORD(lParam)-sc->XL)/(double)sc->X;
                    app.flags.rel_y = (HIWORD(lParam)-sc->YL)/(double)sc->Y;
                    sc->refmark();
                }
                return (0);
            }
        }
        break;

    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
        if (hwnd == app.scr.da) {
            screen *sc = &app.scr;
            if (sc->draw_ghost) {
                if (!sc->first_ghost) {
                    sc->marker(sc->last_x, sc->last_y);
                    sc->first_ghost = true;
                }
                sc->refmark();
                app.flags.rel_x = -1;
                app.flags.rel_y = -1;
                return (0);
            }
        }
        break;
    }
    return (DefWindowProc(hwnd, message, wParam, lParam));
}

#define MSW_FIXED_FONT "Lucida Console"
//#define MSW_FIXED_FONT "Andale Mono"
#define MSW_FIXED_FONT_HEIGHT 12
#define MSW_FIXED_FONT_WIDTH  6

static HFONT
msw_GetFixedFont()
{
    static HFONT fixed;
    if (!fixed) {
        LOGFONT lf;
        lf.lfHeight = MSW_FIXED_FONT_HEIGHT;
        lf.lfWidth = MSW_FIXED_FONT_WIDTH;
        lf.lfEscapement = 0;
        lf.lfOrientation = 0;
        lf.lfWeight = FW_DONTCARE;
        lf.lfItalic = 0;
        lf.lfUnderline = 0;
        lf.lfStrikeOut = 0;
        lf.lfCharSet = ANSI_CHARSET;
        lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
        lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
        lf.lfQuality = 0;
        lf.lfPitchAndFamily = FIXED_PITCH | FF_DONTCARE;
        strcpy(lf.lfFaceName, MSW_FIXED_FONT);
        fixed = CreateFontIndirect(&lf);
    }
    return (fixed);
}


// Create the bitmap struct and set the data field of a new device
// independent bitmap, from the supplied xpm data.  Return true if
// successful
//
static bool
xpm2dib(const char **xpm, BITMAPINFOHEADER **bmp, unsigned char **data)
{
    struct plane { char c; COLORREF cref; };
    if (xpm == 0)
        return (false);
    int nx, ny, nc, xx;
    if (sscanf(xpm[0], "%d %d %d %d", &nx, &ny, &nc, &xx) != 4)
        return (false);
    if (nc < 1)
        return (false);
    plane *pl = new plane[nc];
    for (int i = 0; i < nc; i++) {
        const char *s = xpm[i+1];
        pl[i].c = *s++;
        while (isspace(*s))
            s++;
        while (*s && !isspace(*s))
            s++;
        while (isspace(*s))
            s++;
        if (!strcmp(s, "none")) {
            // assume that bitmap is used in a menu
            pl[i].cref = GetSysColor(COLOR_MENU);
        }
        else if (*s == '#') {
            // used specified rgb
            s++;
            char tbf[4];
            tbf[2] = 0;
            tbf[0] = *s++;
            tbf[1] = *s++;
            int r;
            sscanf(tbf, "%x", &r);
            tbf[0] = *s++;
            tbf[1] = *s++;
            int g;
            sscanf(tbf, "%x", &g);
            tbf[0] = *s++;
            tbf[1] = *s++;
            int b;
            sscanf(tbf, "%x", &b);
            pl[i].cref = RGB(r, g, b);

        }
        else {
            // get the color named
            /*
            int pix = GRX->NameColor(s);
            int r, g, b;
            GRX->RGBofPixel(pix, &r, &g, &b);
            pl[i].cref = RGB(r, g, b);
            */
            // only blue used
            pl[i].cref = RGB(0, 0, 255);
        }
    }
    int tnx = 4*((nx + 3)/4);
    unsigned char *pixels = new unsigned char[tnx*ny];
    int lastv = 0;
    char lastc = 0;
    for (int i = 0; i < ny; i++) {
        const char *s = xpm[nc + ny - i];
        for (int j = 0; j < nx; j++, s++) {
            if (*s != lastc) {
                for (int k = 0; k < nc; k++) {
                    if (*s == pl[k].c) {
                        lastc = *s;
                        lastv = k;
                        break;
                    }
                }
            }
            pixels[i*tnx + j] = lastv;
        }
    }
    char *stuff = new char[sizeof(BITMAPINFOHEADER) + nc*sizeof(RGBQUAD)];
    BITMAPINFOHEADER *bh = (BITMAPINFOHEADER*)stuff;
    bh->biSize = sizeof(BITMAPINFOHEADER);
    bh->biWidth = nx;
    bh->biHeight = ny;
    bh->biPlanes = 1;
    bh->biBitCount = 8;
    bh->biCompression = BI_RGB;
    bh->biSizeImage = tnx*ny;
    bh->biXPelsPerMeter = 0;
    bh->biYPelsPerMeter = 0;
    bh->biClrUsed = nc;
    bh->biClrImportant = 0;

    RGBQUAD *r = (RGBQUAD*)(stuff + bh->biSize);
    for (int i = 0; i < nc; i++) {
        r->rgbRed = GetRValue(pl[i].cref);
        r->rgbGreen = GetGValue(pl[i].cref);
        r->rgbBlue = GetBValue(pl[i].cref);
        r->rgbReserved = 0;
        r++;
    }
    delete [] pl;
    *bmp = bh;
    *data = pixels;
    return (true);
}


// Return a bitmap handle obtained from the text of an XPM
//
static HBITMAP
msw_DIBfromXPM(HWND hwnd, const char **xpm)
{
    BITMAPINFOHEADER *bh;
    unsigned char *pixels;
    if (!xpm2dib(xpm, &bh, &pixels))
        return (0);
    HDC dc = GetDC(hwnd);
    HBITMAP bitmap = CreateDIBitmap(dc, bh, CBM_INIT, (void*)pixels,
        (BITMAPINFO*)bh, DIB_RGB_COLORS);
    ReleaseDC(hwnd, dc);
    delete bh;
    delete [] pixels;
    return (bitmap);
}


static void
create_entry(HWND shell, int resid, int ypos, spinbtn *sb, const char *label)
{
    HINSTANCE instance = GetWindowInstance(shell);
    HWND gb = CreateWindowEx( 
       0,
       "button",                   // Class name
       label,                      // Child caption
       WS_CHILDWINDOW | WS_VISIBLE | BS_GROUPBOX, // window style
       0,                          // origin
       ypos,                       // ...
       60,                         // size
       34,                         // ...
       shell,                      // owner
       (HMENU)resid++,             // no menu
       instance,                   // Instance handle
       0);                         // user-defined parameters
    SendMessage(gb, WM_SETFONT, (long)msw_GetFixedFont(), false);

    sb->entry = CreateWindowEx( 
       0,
       "edit",                     // Class name
       0,                          // Child caption
       WS_CHILDWINDOW | WS_BORDER | WS_VISIBLE, // window style
       0,                          // origin
       ypos + 14,                  // ...
       40,                         // size
       16,                         // ...
       shell,                      // owner
       (HMENU)resid++,             // no menu
       instance,                   // Instance handle
       0);                         // user-defined parameters
    SendMessage(sb->entry, WM_SETFONT, (long)msw_GetFixedFont(), false);

    sb->ud = CreateWindowEx( 
       0,
       "msctls_updown32",  // Class name
       0,                          // Child caption
       WS_CHILDWINDOW | WS_VISIBLE, // window style
       40,                         // origin
       ypos + 12,                  // ...
       10,                         // size
       20,                         // ...
       shell,                      // owner
       (HMENU)resid++,             // no menu
       instance,                   // Instance handle
       0);                         // user-defined parameters

    SendMessage(sb->ud, UDM_SETRANGE, 0,
        MAKELONG((short)32767, (short)-32767));
    SendMessage(sb->ud, UDM_SETPOS, 0, 0);
}


#else

static void
cancel_proc(GtkWidget*, void*)
{
    gtk_signal_disconnect_by_func(GTK_OBJECT(app.shell),
        GTK_SIGNAL_FUNC(cancel_proc), 0);
    gtk_widget_destroy(GTK_WIDGET(app.shell));
    exit(0);
}


static void
help_proc(GtkWidget*, void*)
{
    app.text_display_popup(help_text);
}


static int
da1_expose_hdlr(GtkWidget*, GdkEvent *ev, void*)
{
    if (!app.scr.bpm)
        app.scr.displ();
    else {
        app.scr.refmark();
        app.scr.update(ev->expose.area.x, ev->expose.area.y,
            ev->expose.area.width, ev->expose.area.height);
        app.scr.refmark();
    }
    return (1);
}


static int
da2_expose_hdlr(GtkWidget*, GdkEvent*, void*)
{
    app.diag.diagram();
    return (1);
}


// Motion handler, show marker in marker mode
//
static int
motion_hdlr(GtkWidget*, GdkEvent *event, void *client_data)
{
    screen *sc = (screen*)client_data;
    GdkEventMotion *mev = (GdkEventMotion*)event;  
    if (sc->draw_ghost) {
        if (!sc->first_ghost)
            sc->marker(sc->last_x, sc->last_y);
        sc->last_x = (int)mev->x;
        sc->last_y = (int)mev->y;
        sc->marker(sc->last_x, sc->last_y);
        sc->first_ghost = false;
    }
    return (true);
}


// Leave handler, turn of marker in marker mode
//
static int
leave_hdlr(GtkWidget*, GdkEvent*, void *client_data)
{
    screen *sc = (screen*)client_data;
    if (sc->draw_ghost && !sc->first_ghost)
        sc->marker(sc->last_x, sc->last_y);
    sc->first_ghost = true;
    return (true);
}


// Button handler, set reference in marker mode
//
static int
button_dn_hdlr(GtkWidget*, GdkEvent *event, void *client_data)
{
    screen *sc = (screen*)client_data;
    if (sc->draw_ghost) {
        if (event->button.button == 1) {
            if (event->button.x >= sc->XL && event->button.x <= sc->XU &&
                    event->button.y >= sc->YL && event->button.y <= sc->YU) {
                if (!sc->first_ghost) {
                    sc->marker(sc->last_x, sc->last_y);
                    sc->first_ghost = true;
                }
                sc->refmark();
                app.flags.rel_x = (event->button.x-sc->XL)/sc->X;
                app.flags.rel_y = (event->button.y-sc->YL)/sc->Y;
                sc->refmark();
            }
        }
        if (event->button.button == 2 || event->button.button == 3) {
            if (!sc->first_ghost) {
                sc->marker(sc->last_x, sc->last_y);
                sc->first_ghost = true;
            }
            sc->refmark();
            app.flags.rel_x = -1;
            app.flags.rel_y = -1;
        }
        return (true);
    }
    return (false);
}


// Key press handler, exit marker mode
//
static int
key_dn_hdlr(GtkWidget*, GdkEvent*, void *client_data)
{
    screen *sc = (screen*)client_data;
    if (sc->draw_ghost) {
        sc->refmark();
        if (!sc->first_ghost)
            sc->marker(sc->last_x, sc->last_y);
        sc->draw_ghost = false;
        app.flags.rel_x = -1;
        app.flags.rel_y = -1;
    }
    return (false);
}


static void
wr_btn_hdlr(GtkWidget*, void*)
{
    // This cycles through the X-scale values
    //
    static int num = 2;
    num++;
    if (num > 4)
        num = 1;
    app.Scale = 1 << (num - 1);
    app.scr.displ();
}


static void
menu_hdlr(GtkWidget *caller, void*, unsigned int action)
{
    switch (action) {
    case CODE_DLOG_SAVE:
        app.dialog_popup(DLOG_SAVE);
        break;
    case CODE_DLOG_DEL:
        app.dialog_popup(DLOG_DEL);
        break;
    case CODE_DLOG_LOAD:
        app.dialog_popup(DLOG_LOAD);
        break;
    case CODE_PRINT:
        app.print_popup();
        break;
    case CODE_THREEJ:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.threej = true;
        else
            app.flags.threej = false;
        app.scr.displ();
        app.diag.diagram();
        break;
    case CODE_PHASE:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.z = true;
        else
            app.flags.z = false;
        app.scr.displ();
        break;
    case CODE_X:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.x = true;
        else
            app.flags.x = false;
        app.scr.displ();
        break;
    case CODE_Y:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.y = true;
        else
            app.flags.y = false;
        app.scr.displ();
        break;
    case CODE_AXES:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.s = true;
        else
            app.flags.s = false;
        app.scr.displ();
        break;
    case CODE_SECOND:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.n = true;
        else
            app.flags.n = false;
        app.scr.displ();
        break;
    case CODE_TERT:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.r = true;
        else
            app.flags.r = false;
        app.scr.displ();
        break;
    case CODE_FAIL:
        if (GTK_CHECK_MENU_ITEM(caller)->active)
            app.flags.f = true;
        else
            app.flags.f = false;
        if (app.pts)
            app.scr.displ();
        break;
    case CODE_MARK:
        if (!app.scr.draw_ghost)
            app.scr.xy_pick();
        break;
    }
}


static bool redraw_pending;

static int
redraw_idle_proc(void*)
{
    GdkModifierType state;
    gdk_window_get_pointer(app.shell->window, 0, 0, &state);
    if (state & GDK_BUTTON1_MASK)
        return (1);
    redraw_pending = false;
    app.scr.displ();
    return (0);
}


static void
queue_redraw()
{
    if (!redraw_pending) {
        redraw_pending = true;
        gtk_idle_add(redraw_idle_proc, 0);
    }
}


static void
value_changed(GtkWidget *widget, void *client_data)
{
    const char *text = gtk_entry_get_text(GTK_ENTRY(widget));
    if (!text || !*text)
        return;
    switch ((long)client_data) {
    default:
        return;
    case VAL_L1:
        vals.la = atof(text);
        break;
    case VAL_L2:
        vals.lb = atof(text);
        break;
    case VAL_L3:
        vals.lc = atof(text);
        break;
    case VAL_L4:
        vals.ld = atof(text);
        break;
    case VAL_FRAC:
        vals.lfrac = atof(text);
        break;
    case VAL_IC1:
        vals.ix = atof(text);
        break;
    case VAL_IC2:
        vals.i2 = atof(text);
        break;
    case VAL_IC3:
        vals.iy = atof(text);
        break;
    case VAL_COUP:
        app.Coupling = atof(text);
        break;
    }
    queue_redraw();
}


// Respond to a button-press by posting a menu passed in as widget.
//
// Note that the "widget" argument is the menu being posted, NOT
// the button that was pressed.
//
static int
button_press(GtkWidget *widget, GdkEvent *event)
{
    if (event->type == GDK_BUTTON_PRESS) {
        GdkEventButton *bevent = (GdkEventButton*)event;
        gtk_menu_popup (GTK_MENU(widget), 0, 0, 0, 0, bevent->button,
            bevent->time);
        // Tell calling code that we have handled this event; the buck
        // stops here.
        return (true);
    }
            
    // Tell calling code that we have not handled this event; pass it on.
    return (false);
}


static void
recall_menu_hdlr(GtkWidget *menu_item, void*)
{
    char *name = (char*)gtk_object_get_data(GTK_OBJECT(menu_item), "name");
    if (name)
        app.recall(name);
}

#endif


//
// The printer interface
//

#ifdef WIN32_NATIVE
static BOOL CALLBACK printDlgProc(HWND, UINT, WPARAM, LPARAM);
void client_rect(HWND, HWND, RECT*);
static void print_resize(HWND);
#else
static void tofile_proc(GtkWidget*, void*);
static void go_proc(GtkWidget*, void*);
static void print_cancel_proc(GtkWidget*, void*);
#ifdef WIN32
static void print_menu_proc(GtkWidget*, void*);
#endif
#endif


void
panel::print_popup()
{
    if (!cmdtext) {
#ifdef WIN32
#else
        if (!PRINTCMD)
            cmdtext = strdup("lpr -h");
        else
            cmdtext = strdup(PRINTCMD);
#endif
    }
    if (!tofilename)
        tofilename = strdup(OUTNAME);
#ifdef WIN32_NATIVE
    HWND hwnd = CreateDialog(GetWindowInstance(shell),
        MAKEINTRESOURCE(IDD_PRINT), shell, printDlgProc);
    if (!hwnd)
        return;
    HWND label = GetDlgItem(hwnd, IDC_PRINT_STATIC);
    Static_SetText(label, tofile ? "File Name:" : "Printer Name:");

    HWND button = GetDlgItem(hwnd, IDC_PRINT_BTN);
    Button_SetCheck(button, tofile);

    HWND cbox = GetDlgItem(hwnd, IDC_PRINT_CBOX);
    int curpr;
    char **prlist;
    int numpr = msw_ListPrinters(&curpr, &prlist);
    if (numpr > 0) {
        for (int i = 0; i < numpr; i++) {
            ComboBox_AddString(cbox, prlist[i]);
        }
        ComboBox_SetCurSel(cbox, curpr);
        cmdtext = strdup(prlist[curpr]);
        for (int i = 0; i < numpr; i++)
            delete [] prlist[i];
        delete [] prlist;
    }
    else {
        ComboBox_AddString(cbox, "no printers found");
        cmdtext = strdup("default");
    }

    HWND edit = GetDlgItem(hwnd, IDC_PRINT_EDIT);
    Edit_SetText(edit, tofilename);
    if (tofile)
        ShowWindow(cbox, SW_HIDE);
    else
        ShowWindow(edit, SW_HIDE);

    RECT r;
    GetWindowRect(hwnd, &r);
    int wid = r.right - r.left;
    int hei = r.bottom - r.top;
    GetWindowRect(shell, &r);
    MoveWindow(hwnd, r.left + 100, r.top + 100, wid, hei, false);

    ShowWindow(hwnd, SW_SHOW);

#else
#if GTK_CHECK_VERSION(2,0,0)
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#else
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_DIALOG);
#endif
    gtk_window_set_title(GTK_WINDOW(popup), "PostScript Print");
    gtk_signal_connect(GTK_OBJECT(popup), "destroy",
        GTK_SIGNAL_FUNC(print_cancel_proc), popup);

    GtkWidget *form = gtk_table_new(1, 3, false);
    gtk_widget_show(form);
    gtk_container_add(GTK_CONTAINER(popup), form);
    gtk_container_set_border_width(GTK_CONTAINER(popup), 2);

    //
    // framed printer command label and entry area
    //
    GtkWidget *row2 = gtk_hbox_new(false, 2);
    gtk_widget_show(row2);

    GtkWidget *cmdlab = gtk_label_new(tofile ? "File Name" : "Print Command");
    gtk_widget_show(cmdlab);
    gtk_misc_set_alignment(GTK_MISC(cmdlab), 0, 0.5);
    gtk_misc_set_padding(GTK_MISC(cmdlab), 2, 2);
    gtk_box_pack_start(GTK_BOX(row2), cmdlab, true, true, 0);

    GtkWidget *button = gtk_check_button_new_with_label("To File");
    gtk_widget_show(button);
    if (tofile)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), true);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
        GTK_SIGNAL_FUNC(tofile_proc), 0);
    gtk_box_pack_end(GTK_BOX(row2), button, false, false, 0);

    GtkWidget *frame = gtk_frame_new(0);
    gtk_widget_show(frame);
    gtk_container_add(GTK_CONTAINER(frame), row2);
    row2 = gtk_hbox_new(false, 2);
    gtk_widget_show(row2);
    gtk_box_pack_start(GTK_BOX(row2), frame, true, true, 0);
    int rowcnt = 0;
    gtk_table_attach(GTK_TABLE(form), row2, 0, 1, rowcnt, rowcnt+1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 2, 2);
    rowcnt++;

#ifdef WIN32
    // In Windows, we have a list of printers when not printing to
    // a file.
    GtkWidget *optmenu = gtk_option_menu_new();
    gtk_widget_show(optmenu);
    GtkWidget *menu = gtk_menu_new();
    gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), menu);
    int curpr;
    char **prlist;
    int numpr = msw_ListPrinters(&curpr, &prlist);
    if (numpr > 0) {
        for (int i = 0; i < numpr; i++) {
            GtkWidget *mi = gtk_menu_item_new_with_label(prlist[i]);
            gtk_widget_set_name(mi, prlist[i]);
            gtk_widget_show(mi);
            gtk_menu_append(GTK_MENU(menu), mi);
            gtk_signal_connect(GTK_OBJECT(mi), "activate",
                GTK_SIGNAL_FUNC(print_menu_proc), this);
        }
        gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), curpr);
        cmdtext = strdup(prlist[curpr]);
        for (int i = 0; i < numpr; i++)
            delete [] prlist[i];
        delete [] prlist;
    }
    else {
        GtkWidget *mi = gtk_menu_item_new_with_label("no printers found");
        gtk_widget_show(mi);
        gtk_menu_append(GTK_MENU(menu), mi);
        cmdtext = strdup("default");
    }
    gtk_table_attach(GTK_TABLE(form), optmenu, 0, 1, rowcnt, rowcnt+1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);
    rowcnt++;
#endif

    GtkWidget *cmdtxtbox = gtk_entry_new();
    gtk_widget_show(cmdtxtbox);
#ifdef WIN32
    gtk_entry_set_text(GTK_ENTRY(cmdtxtbox), tofilename);
#else
    gtk_entry_set_text(GTK_ENTRY(cmdtxtbox),
        tofile ? tofilename : cmdtext);
#endif
    gtk_widget_set_usize(cmdtxtbox, 200, -1);
    gtk_object_set_data(GTK_OBJECT(button), "cmdlab", cmdlab);
    gtk_object_set_data(GTK_OBJECT(button), "cmdtxtbox", cmdtxtbox);

    gtk_table_attach(GTK_TABLE(form), cmdtxtbox, 0, 1, rowcnt, rowcnt+1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);
    rowcnt++;

#ifdef WIN32
    gtk_object_set_data(GTK_OBJECT(button), "optmenu", optmenu);
    if (tofile)
        gtk_widget_hide(optmenu);
    else
        gtk_widget_hide(cmdtxtbox);
#endif

    GtkWidget *sep = gtk_hseparator_new();
    gtk_widget_show(sep);
    gtk_table_attach(GTK_TABLE(form), sep, 0, 1, rowcnt, rowcnt+1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);
    rowcnt++;

    //
    // print and dismiss buttons
    //
    GtkWidget *row5 = gtk_hbox_new(false, 2);
    gtk_widget_show(row5);

    button = gtk_button_new_with_label("Print");
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
        GTK_SIGNAL_FUNC(go_proc), popup);
    gtk_box_pack_start(GTK_BOX(row5), button, true, true, 0);

    button = gtk_button_new_with_label("Dismiss");
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
        GTK_SIGNAL_FUNC(print_cancel_proc), popup);
    gtk_box_pack_start(GTK_BOX(row5), button, true, true, 0);

    gtk_table_attach(GTK_TABLE(form), row5, 0, 1, rowcnt, rowcnt+1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);

    int x, y;
    gdk_window_get_root_origin(scr.da->window, &x, &y);
    gtk_widget_set_uposition(popup, x + 150, y + 100);
    gtk_window_set_transient_for(GTK_WINDOW(popup), GTK_WINDOW(shell));
    gtk_widget_show(popup);
#endif
}


#ifdef WIN32_NATIVE

static BOOL CALLBACK
printDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    char buf[256];
    switch (message) {
    case WM_INITDIALOG:
        return (true);

    case WM_ACTIVATE:
        return (true);

    case WM_SIZE:
        print_resize(hwnd);
        return (true);

    case WM_COMMAND:
        // handle button presses
        if (LOWORD(wParam) == IDOK) {
            if (!app.tofile) {
                char tmpbuf[64];
                strcpy(tmpbuf, "/tmp/iphcXXXXXX");
                char *tmp = app.tofilename;
                app.tofilename = mktemp(tmpbuf);
                app.printscreen();
                app.tofilename = tmp;
            }
            else
                app.printscreen();
            DestroyWindow(hwnd);
            return (true);
        }
        if (LOWORD(wParam) == IDC_PRINT_BTN) {
            bool state =
                !((Button_GetState((HWND)lParam) & 0x3) == BST_UNCHECKED);
            app.tofile = state;
            HWND cmdlab = GetDlgItem(hwnd, IDC_PRINT_STATIC);
            HWND cmdtxtbox = GetDlgItem(hwnd, IDC_PRINT_EDIT);
            HWND cmbbox = GetDlgItem(hwnd, IDC_PRINT_CBOX);
            if (cmdlab && cmdtxtbox && cmbbox) {
                if (state) {
                    Static_SetText(cmdlab, "File Name");
                    ShowWindow(cmbbox, SW_HIDE);
                    ShowWindow(cmdtxtbox, SW_SHOW);
                }
                else {
                    Static_SetText(cmdlab, "Print Command");
                    ShowWindow(cmdtxtbox, SW_HIDE);
                    ShowWindow(cmbbox, SW_SHOW);
                } 
            }
            return (true);
        }
        if (LOWORD(wParam) == IDC_PRINT_CBOX) {
            if (HIWORD(wParam) == CBN_SELENDOK) {
                ComboBox_GetText((HWND)lParam, buf, 256);
                delete [] app.cmdtext;
                app.cmdtext = strdup(buf);
            }
            return (true);
        }
        if (LOWORD(wParam) != IDCANCEL)
            return (false);
        // fall thru

    case WM_CLOSE:
        DestroyWindow(hwnd);
        return (true);
    }
    return (false);
}


void
client_rect(HWND dlg, HWND ctrl, RECT *r)
{
    GetWindowRect(ctrl, r);
    ScreenToClient(dlg, (LPPOINT)&r->left);
    ScreenToClient(dlg, (LPPOINT)&r->right);
}


static void
print_resize(HWND hwnd)
{
    RECT rwin;
    GetClientRect(hwnd, &rwin);

    RECT rc;
    HWND grp = GetDlgItem(hwnd, IDC_PRINT_GROUP);
    client_rect(hwnd, grp, &rc);
    int marg = rc.left - rwin.left;
    rc.right = rwin.right - marg;
    MoveWindow(grp, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    HWND btn = GetDlgItem(hwnd, IDC_PRINT_BTN);
    client_rect(hwnd, btn, &rc);
    rc.left = rwin.right - 2*marg - (rc.right - rc.left);
    rc.right = rwin.right - 2*marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);
    int bp = rc.left - marg;

    HWND label = GetDlgItem(hwnd, IDC_PRINT_STATIC);
    client_rect(hwnd, label, &rc);
    rc.right = bp;
    MoveWindow(label, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    HWND text = GetDlgItem(hwnd, IDC_PRINT_EDIT);
    client_rect(hwnd, text, &rc);
    rc.right = rwin.right - marg;
    MoveWindow(text, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    int bw = (rwin.right - rwin.left - 3*marg)/2;

    btn = GetDlgItem(hwnd, IDOK);
    client_rect(hwnd, btn, &rc);
    rc.right = rc.left + bw;
    rc.bottom = rwin.bottom - marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);
    int bpos = rc.right + marg;

    btn = GetDlgItem(hwnd, IDCANCEL);
    client_rect(hwnd, btn, &rc);
    rc.left = bpos;
    rc.right = rwin.right - marg;
    rc.bottom = rwin.bottom - marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    InvalidateRect(hwnd, 0, true);    
}


#else

static void
tofile_proc(GtkWidget *caller, void*)
{
    bool state = (GTK_TOGGLE_BUTTON(caller)->active ? true : false);
    app.tofile = state;
    GtkWidget *cmdlab = (GtkWidget*)gtk_object_get_data(GTK_OBJECT(caller),
        "cmdlab");
    GtkWidget *cmdtxtbox = (GtkWidget*)gtk_object_get_data(GTK_OBJECT(caller),
        "cmdtxtbox");
#ifdef WIN32
    GtkWidget *optmenu = (GtkWidget*)gtk_object_get_data(GTK_OBJECT(caller),
        "optmenu");
    if (cmdlab && cmdtxtbox && optmenu) {
        if (state) {
            gtk_label_set_text(GTK_LABEL(cmdlab), "File Name");
            gtk_widget_hide(optmenu);
            gtk_widget_show(cmdtxtbox);
        }
        else {
            gtk_label_set_text(GTK_LABEL(cmdlab), "Print Command");
            gtk_widget_hide(cmdtxtbox);
            gtk_widget_show(optmenu);
        }
    }
#else
    if (cmdlab && cmdtxtbox) {
        const char *s = gtk_entry_get_text(GTK_ENTRY(cmdtxtbox));
        if (state) {
            gtk_label_set_text(GTK_LABEL(cmdlab), "File Name");
            if (app.cmdtext)
                delete [] app.cmdtext;
            app.cmdtext = strdup(s);
            gtk_entry_set_text(GTK_ENTRY(cmdtxtbox), app.tofilename);
        }
        else {
            gtk_label_set_text(GTK_LABEL(cmdlab), "Print Command");
            if (app.tofilename)
                delete [] app.tofilename;
            app.tofilename = strdup(s);
            gtk_entry_set_text(GTK_ENTRY(cmdtxtbox), app.cmdtext);
        } 
    }
#endif
}


static void
go_proc(GtkWidget*, void *popup)
{
    if (!app.tofile) {
        char tmpbuf[64];
        strcpy(tmpbuf, "/tmp/iphcXXXXXX");
        char *tmp = app.tofilename;
#ifdef HAVE_MKSTEMP
        int fd = mkstemp(tmpbuf);
        close(fd);
        app.tofilename = tmpbuf;
#else
        app.tofilename = mktemp(tmpbuf);
#endif
        app.printscreen();
        app.tofilename = tmp;
    }
    else
        app.printscreen();
    print_cancel_proc(0, popup);
}


static void
print_cancel_proc(GtkWidget*, void *popup)
{
    gtk_signal_disconnect_by_func(GTK_OBJECT(popup),
        GTK_SIGNAL_FUNC(print_cancel_proc), popup);
    gtk_widget_destroy(GTK_WIDGET(popup));
}


#ifdef WIN32

static void
print_menu_proc(GtkWidget *w, void *arg)
{
    panel *p = (panel*)arg;
    delete [] p->cmdtext;
    p->cmdtext = strdup(gtk_widget_get_name(w));
}
#endif
#endif


//
// Dialog for Save/Delete
//

#ifdef WIN32_NATIVE
static BOOL CALLBACK textDlgProc(HWND, UINT, WPARAM, LPARAM);
static void dialog_resize(HWND);
#else
static void dialog_cancel(GtkWidget*, void*);
static void dialog_save_action(GtkWidget*, void*);
static void dialog_del_action(GtkWidget*, void*);
static void dialog_load_action(GtkWidget*, void*);
#endif


void
panel::dialog_popup(unsigned int action_arg)
{
#ifdef WIN32_NATIVE
    HWND hwnd = CreateDialog(GetWindowInstance(shell),
        MAKEINTRESOURCE(IDD_TEXTPOP), shell, textDlgProc);
    if (!hwnd)
        return;
    HWND label = GetDlgItem(hwnd, IDC_TEXTPOP_STATIC);
    if (action_arg == DLOG_LOAD)
        Static_SetText(label, "Enter Filename:");
    else
        Static_SetText(label, "Enter Dataset Name:");

    HWND button = GetDlgItem(hwnd, IDOK);
    const char *action_str = "Go";
    if (action_arg == DLOG_SAVE)
        action_str = "Save Dataset";
    else if (action_arg == DLOG_DEL)
        action_str = "Delete Dataset";
    else if (action_arg == DLOG_LOAD)
        action_str = "Load Points";
    Button_SetText(button, action_str);

    RECT r;
    GetWindowRect(hwnd, &r);
    int wid = r.right - r.left;
    int hei = r.bottom - r.top;
    GetWindowRect(shell, &r);
    MoveWindow(hwnd, r.left + 100, r.top + 100, wid, hei, false);

    SetWindowLong(hwnd, GWL_USERDATA, action_arg);
    ShowWindow(hwnd, SW_SHOW);

#else
#if GTK_CHECK_VERSION(2,0,0)
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#else
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_DIALOG);
#endif
    gtk_window_set_title(GTK_WINDOW(popup), "Dialog");
    gtk_signal_connect(GTK_OBJECT(popup), "destroy",
        GTK_SIGNAL_FUNC(dialog_cancel), popup);
    
    GtkWidget *form = gtk_table_new(1, 3, false);
    gtk_widget_show(form);
    gtk_container_add(GTK_CONTAINER(popup), form);

    //
    // the label, in a frame
    //
    GtkWidget *frame = gtk_frame_new(0);
    gtk_widget_show(frame);
    GtkWidget *label = gtk_label_new(
        action_arg == DLOG_LOAD ? "Enter filename:" : "Enter Dataset Name:");
    gtk_widget_show(label);
    gtk_misc_set_padding(GTK_MISC(label), 4, 2);
    gtk_container_add(GTK_CONTAINER(frame), label);
    gtk_table_attach(GTK_TABLE(form), frame, 0, 1, 0, 1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 2, 2);

    //
    // the entry area
    //
    GtkWidget *entry = gtk_entry_new();
    gtk_widget_show(entry);
    gtk_object_set_data(GTK_OBJECT(entry), "popup", popup);
    gtk_object_set_data(GTK_OBJECT(entry), "label", label);
    gtk_table_attach(GTK_TABLE(form), entry, 0, 1, 1, 2,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 0);

    int wid = 160;
    gtk_widget_set_usize(entry, wid + 10, -1);
    //
    // the button line, action and dismiss
    //
    const char *action_str = "Go";
    if (action_arg == DLOG_SAVE)
        action_str = "Save Dataset";
    else if (action_arg == DLOG_DEL)
        action_str = "Delete Dataset";
    else if (action_arg == DLOG_LOAD)
        action_str = "Load Points";
    GtkWidget *hbox = gtk_hbox_new(false, 2);
    gtk_widget_show(hbox);
    GtkWidget *action = gtk_button_new_with_label(action_str);
    gtk_widget_show(action);
    // info needed for callback
    gtk_box_pack_start(GTK_BOX(hbox), action, true, true, 0);
    if (action_arg == DLOG_SAVE)
        gtk_signal_connect(GTK_OBJECT(action), "clicked",
            GTK_SIGNAL_FUNC(dialog_save_action), entry);
    else if (action_arg == DLOG_DEL)
        gtk_signal_connect(GTK_OBJECT(action), "clicked",
            GTK_SIGNAL_FUNC(dialog_del_action), entry);
    else if (action_arg == DLOG_LOAD)
        gtk_signal_connect(GTK_OBJECT(action), "clicked",
            GTK_SIGNAL_FUNC(dialog_load_action), entry);

    GtkWidget *cancel = gtk_button_new_with_label("Cancel");
    gtk_widget_show(cancel);
    gtk_box_pack_start(GTK_BOX(hbox), cancel, true, true, 0);
    gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
        GTK_SIGNAL_FUNC(dialog_cancel), popup);

    gtk_table_attach(GTK_TABLE(form), hbox, 0, 1, 2, 3,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);

    int x, y;
    gdk_window_get_root_origin(scr.da->window, &x, &y);
    gtk_widget_set_uposition(popup, x + 150, y + 100);
    gtk_window_set_transient_for(GTK_WINDOW(popup), GTK_WINDOW(shell));
    gtk_widget_show(popup);
#endif
}


#ifdef WIN32_NATIVE

static BOOL CALLBACK
textDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
    case WM_INITDIALOG:
        return (true);

    case WM_ACTIVATE:
        return (true);

    case WM_SIZE:
        dialog_resize(hwnd);
        return (true);

    case WM_COMMAND:
        // handle button presses
        if (LOWORD(wParam) == IDOK) {
            long action_arg = GetWindowLong(hwnd, GWL_USERDATA);
            char buf[256];
            HWND edit = GetDlgItem(hwnd, IDC_TEXTPOP_EDIT);
            if (action_arg == DLOG_SAVE) {
                Edit_GetText(edit, buf, 256);
                if (*buf && !app.save(buf, false))
                    DestroyWindow(hwnd);
                else {
                    HWND label = GetDlgItem(hwnd, IDC_TEXTPOP_STATIC);
                    if (label)
                        Static_SetText(label, "Bad entry, try again");
                }
            }
            else if (action_arg == DLOG_DEL) {
                Edit_GetText(edit, buf, 256);
                if (*buf && !app.save(buf, true))
                    DestroyWindow(hwnd);
                else {
                    HWND label = GetDlgItem(hwnd, IDC_TEXTPOP_STATIC);
                    if (label)
                        Static_SetText(label, "Not found, try again");
                }
            }
            else if (action_arg == DLOG_LOAD) {
                Edit_GetText(edit, buf, 256);
                if (*buf) {
                    FILE *fp = fopen(buf, "r");
                    if (fp) {
                        app.extpts(fp);
                        app.scr.plpts();
                        fclose(fp);
                        DestroyWindow(hwnd);
                        return (true);
                    }
                }
                else {
                    HWND label = GetDlgItem(hwnd, IDC_TEXTPOP_STATIC);
                    if (label)
                        Static_SetText(label, "Can't open, try again");
                }
            }
            return (true);
        }
        else if (LOWORD(wParam) != IDCANCEL)
            return (false);
        // fall thru

    case WM_CLOSE:
        DestroyWindow(hwnd);
        return (true);
    }
    return (false);
}


static void
dialog_resize(HWND hwnd)
{
    RECT rwin;
    GetClientRect(hwnd, &rwin);

    RECT rc;
    HWND grp = GetDlgItem(hwnd, IDC_TEXTPOP_GROUP);
    client_rect(hwnd, grp, &rc);
    int marg = rc.left - rwin.left;
    rc.right = rwin.right - marg;
    MoveWindow(grp, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    HWND label = GetDlgItem(hwnd, IDC_TEXTPOP_STATIC);
    client_rect(hwnd, label, &rc);
    rc.right = rwin.right - 2*marg;
    MoveWindow(label, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    HWND text = GetDlgItem(hwnd, IDC_TEXTPOP_EDIT);
    client_rect(hwnd, text, &rc);
    rc.right = rwin.right - marg;
    MoveWindow(text, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    int bw = (rwin.right - rwin.left - 3*marg)/2;

    HWND btn = GetDlgItem(hwnd, IDOK);
    client_rect(hwnd, btn, &rc);
    rc.right = rc.left + bw;
    rc.bottom = rwin.bottom - marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);
    int bpos = rc.right + marg;

    btn = GetDlgItem(hwnd, IDCANCEL);
    client_rect(hwnd, btn, &rc);
    rc.left = bpos;
    rc.right = rwin.right - marg;
    rc.bottom = rwin.bottom - marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    InvalidateRect(hwnd, 0, true);    
}


#else

static void
dialog_cancel(GtkWidget*, void *popup)
{
    gtk_signal_disconnect_by_func(GTK_OBJECT(popup),
        GTK_SIGNAL_FUNC(dialog_cancel), popup);
    gtk_widget_destroy(GTK_WIDGET(popup));
}


// Dialog action to save a data set
//
static void
dialog_save_action(GtkWidget*, void *entry)
{
    const char *s = gtk_entry_get_text(GTK_ENTRY(entry));
    if (s && *s && !app.save(s, false)) {
        GtkWidget *popup =
            (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "popup");
        if (popup)
            dialog_cancel(0, popup);
        return;
    }
    GtkWidget *label =
        (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "label");
    if (label)
        gtk_label_set_text(GTK_LABEL(label), "Bad entry, try again");
}


// Dialog action to delete a data set
//
static void
dialog_del_action(GtkWidget*, void *entry)
{
    const char *s = gtk_entry_get_text(GTK_ENTRY(entry));
    if (s && *s && !app.save(s, true)) {
        GtkWidget *popup =
            (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "popup");
        if (popup)
            dialog_cancel(0, popup);
        return;
    }
    GtkWidget *label =
        (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "label");
    if (label)
        gtk_label_set_text(GTK_LABEL(label), "Not found, try again");
}


// Dialog action to load a "points" file
//
static void
dialog_load_action(GtkWidget*, void *entry)
{
    const char *s = gtk_entry_get_text(GTK_ENTRY(entry));
    if (s && *s) {
        FILE *fp = fopen(s, "r");
        if (fp) {
            app.extpts(fp);
            app.scr.plpts();
            fclose(fp);
            GtkWidget *popup =
                (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "popup");
            if (popup)
                dialog_cancel(0, popup);
            return;
        }
    }
    GtkWidget *label =
        (GtkWidget*)gtk_object_get_data(GTK_OBJECT(entry), "label");
    if (label)
        gtk_label_set_text(GTK_LABEL(label), "Can't open, try again");
}

#endif


//
// Popup to display the help text
//

#ifdef WIN32_NATIVE
static BOOL CALLBACK helpDlgProc(HWND, UINT, WPARAM, LPARAM);
static void text_resize(HWND);
#else
static void text_cancel(GtkWidget*, void*);
#endif


void
panel::text_display_popup(char *message_str)
{
#ifdef WIN32_NATIVE
    HWND hwnd = CreateDialog(GetWindowInstance(shell),
        MAKEINTRESOURCE(IDD_HELP), shell, helpDlgProc);
    if (!hwnd)
        return;

    HWND text = GetDlgItem(hwnd, IDC_HELP_Help);
    Edit_SetText(text, help_text);

    RECT r;
    GetWindowRect(hwnd, &r);
    int wid = r.right - r.left;
    int hei = r.bottom - r.top;
    GetWindowRect(shell, &r);
    MoveWindow(hwnd, r.left + 100, r.top + 100, wid, hei, false);

    ShowWindow(hwnd, SW_SHOW);

#else
#if GTK_CHECK_VERSION(2,0,0)
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#else
    GtkWidget *popup = gtk_window_new(GTK_WINDOW_DIALOG);
#endif
    gtk_window_set_title(GTK_WINDOW(popup), "Text Display");
    gtk_signal_connect(GTK_OBJECT(popup), "destroy",
        GTK_SIGNAL_FUNC(text_cancel), popup);
    
    GtkWidget *form = gtk_table_new(1, 3, false);
    gtk_widget_show(form);
    gtk_container_set_border_width(GTK_CONTAINER(form), 2);
    gtk_container_add(GTK_CONTAINER(popup), form);

    //
    // scrolled text area
    //
    GtkWidget *hbox = gtk_hbox_new(false, 0);
    gtk_widget_show(hbox);

#if GTK_CHECK_VERSION(2,0,0)
    GtkWidget *swin = gtk_scrolled_window_new(0, 0);
    gtk_widget_show(swin);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
        GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GtkWidget *text = gtk_text_view_new();
    gtk_widget_show(text);
    gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text), GTK_WRAP_NONE);
    gtk_text_view_set_left_margin(GTK_TEXT_VIEW(text), 4);
    gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text), TRUE);  
    gtk_container_add(GTK_CONTAINER(swin), text);
    gtk_box_pack_start(GTK_BOX(hbox), swin, true, true, 0);

    GtkTextBuffer *tbf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text));  
    gtk_text_buffer_set_text(tbf, message_str, -1);
#else
    GtkWidget *text = gtk_text_new(0, 0);
    gtk_widget_show(text);
    gtk_box_pack_start(GTK_BOX(hbox), text, true, true, 0);

    GtkWidget *vscroll = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
    gtk_widget_show(vscroll);
    gtk_box_pack_start(GTK_BOX(hbox), vscroll, false, false, 0);

    gtk_text_insert(GTK_TEXT(text), 0, 0, 0, message_str, -1);
#endif

    gtk_table_attach(GTK_TABLE(form), hbox, 0, 1, 0, 1,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 2, 2);

    gtk_widget_set_usize(text, 500, 300);

    GtkWidget *sep = gtk_hseparator_new();
    gtk_widget_show(sep);
    gtk_table_attach(GTK_TABLE(form), sep, 0, 1, 1, 2,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);

    GtkWidget *button = gtk_button_new_with_label("Dismiss");
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
        GTK_SIGNAL_FUNC(text_cancel), popup);

    gtk_table_attach(GTK_TABLE(form), button, 0, 1, 2, 3,
        (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK),
        (GtkAttachOptions)0, 2, 2);

    int x, y;
    gdk_window_get_root_origin(scr.da->window, &x, &y);
    gtk_widget_set_uposition(popup, x + 150, y + 100);
    gtk_window_set_transient_for(GTK_WINDOW(popup), GTK_WINDOW(shell));
    gtk_widget_show(popup);
#endif
}


#ifdef WIN32_NATIVE

static BOOL CALLBACK
helpDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
    case WM_SIZE:
        text_resize(hwnd);
        return (true);

    case WM_COMMAND:
    case WM_CLOSE:
        DestroyWindow(hwnd);
        return (true);
    }
    return (false);
}


static void
text_resize(HWND hwnd)
{
    RECT rwin;
    GetClientRect(hwnd, &rwin);

    RECT rc;
    HWND btn = GetDlgItem(hwnd, IDCANCEL);
    client_rect(hwnd, btn, &rc);
    int marg = rc.left - rwin.left;
    rc.right = rwin.right - marg;
    rc.top = rwin.bottom - marg - (rc.bottom - rc.top);
    rc.bottom = rwin.bottom - marg;
    MoveWindow(btn, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);
    int bot = rc.top - marg;

    HWND text = GetDlgItem(hwnd, IDC_HELP_Help);
    client_rect(hwnd, text, &rc);
    rc.right = rwin.right - marg;
    rc.bottom = bot;
    MoveWindow(text, rc.left, rc.top, rc.right - rc.left,
        rc.bottom - rc.top, false);

    InvalidateRect(hwnd, 0, true);    
}


#else

static void
text_cancel(GtkWidget*, void *popup)
{
    gtk_signal_disconnect_by_func(GTK_OBJECT(popup),
        GTK_SIGNAL_FUNC(text_cancel), popup);
    gtk_widget_destroy(GTK_WIDGET(popup));
}

#endif


// Extract points from file
//
void
panel::extpts(FILE *fp)
{
    int i = 0;
    char buf[80];
    while (fgets(buf,80,fp)) {
        if (strchr(buf,'p') || strchr(buf,'f')) i++;
    }
    extra *p = new extra;
    p->x = new float[i];
    p->y = new float[i];
    p->pf = new int[i];
    p->size = i;
    p->next = NULL;
    rewind(fp);
    i = 0;
    while (fgets(buf,80,fp)) {
        double v1, v2;
        if (*buf != '\t') sscanf(buf,"%*d %*d %le %le",&v1,&v2);
        else if (strchr(buf,'p')) {
            (p->x)[i] = (float) v1;
            (p->y)[i] = (float) v2;
            (p->pf)[i] = 1;
            i++;
        }
        else if (strchr(buf,'f')) {
            (p->x)[i] = (float) v1;
            (p->y)[i] = (float) v2;
            (p->pf)[i] = 0;
            i++;
        }
    }
    if (!pts) pts = p;
    else {
        extra *q;
        for (q = pts; q->next; q = q->next)
            ;
        q->next = p;
    }
    fclose(fp);
}


// Delete the last set of extracted points
//
void
panel::delpts()
{
    if (!pts) return;
    struct extra *p;
    for (p = pts; p->next && p->next->next; p = p->next)
        ;
    if (p->next) {        
        delete [] p->next->x;
        delete [] p->next->y;
        delete [] p->next->pf;
        delete p->next;
        p->next = 0;
    }
    else {
        delete [] p->x;
        delete [] p->y;
        delete [] p->pf;
        delete p;
        pts = 0;
    }
}


bool
panel::save(const char *name, bool delete_ent)
{
    char tmpbuf[64];
    strcpy(tmpbuf, "/tmp/ipXXXXXX");
#ifdef HAVE_MKSTEMP
    const char *tmpfile = tmpbuf;
    int fd = mkstemp(tmpbuf);
    FILE *gp = fdopen(fd, "w");
#else
    char *tmpfile = mktemp(tmpbuf);
    FILE *gp = fopen(tmpfile, "w");
#endif
    if (gp) {
        char buf[256];
        char *datafile = get_data_file(DATANAME);
        bool update = false;
        FILE *fp = fopen(datafile, "r");
        if (fp) {
            while (fgets(buf, 256, fp) != 0) {
                if (!update && !strncmp(name, buf, strlen(name))) {
                    sprintf(buf, "%s %d %g %g %g %g %g %g %g %g %g\n",
                        name, flags.threej ? 3 : 2, vals.la, vals.lb,
                        vals.lc, vals.ld, vals.lfrac, vals.ix, vals.i2,
                        vals.iy, Coupling);
                    update = true;
                    if (delete_ent)
                        continue;
                }
                fputs(buf, gp);
            }
            fclose(fp);
        }
        else
            fputs("# ipat stored data sets\n", gp);
        if (!update) {
            if (delete_ent) {
                fclose(gp);
                unlink(tmpfile);
                return (true);
            }
            sprintf(buf, "%s %d %g %g %g %g %g %g %g %g %g\n",
                name, flags.threej ? 3 : 2, vals.la, vals.lb,
                vals.lc, vals.ld, vals.lfrac, vals.ix, vals.i2,
                vals.iy, Coupling);
            fputs(buf, gp);
        }
        fclose(gp);
        unlink(datafile);

        fp = fopen(datafile, "w");
        delete [] datafile;
        if (!fp) {
            errmsg(strerror(errno), 0);
            return (true);
        }
        gp = fopen(tmpfile, "r");
        if (!gp) {
            errmsg(strerror(errno), 0);
            fclose(fp);
            return (true);
        }
        while (fgets(buf, 256, gp) != 0)
            fputs(buf, fp);
        fclose(gp);
        fclose(fp);
        unlink(tmpfile);

        init_recall_menu();
        return (false);
    }
    else
        errmsg(strerror(errno), 0);
    return (true);
}


bool
panel::recall(const char *name)
{
    char *datafile = get_data_file(DATANAME);
    FILE *fp = fopen(datafile, "r");
    delete [] datafile;
    if (fp) {
        char buf[256];
        while (fgets(buf, 256, fp) != 0) {
            if (!strncmp(name, buf, strlen(name))) {
                int njj;
                double la, lb, lc, ld, f, ix, i2, iy, c;
                int i = sscanf(buf + strlen(name),
                    "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf",
                    &njj, &la, &lb, &lc, &ld, &f, &ix, &i2, &iy, &c);
                if (i == 10) {
                    flags.threej = njj == 3 ? true : false;
#ifdef WIN32_NATIVE
                    set_status(ID_Options_ThreeJunctions, flags.threej);
                    entries[0].setval(la);
                    vals.la = la;
                    entries[1].setval(lb);
                    vals.lb = lb;
                    entries[2].setval(lc);
                    vals.lc = lc;
                    entries[3].setval(ld);
                    vals.ld = ld;
                    entries[4].setval(f);
                    vals.lfrac = f;
                    entries[5].setval(ix);
                    vals.ix = ix;
                    entries[6].setval(i2);
                    vals.i2 = i2;
                    entries[7].setval(iy);
                    vals.iy = iy;
                    entries[8].setval(c);
                    Coupling = c;
                    InvalidateRect(scr.da, 0, false);
                    InvalidateRect(diag.da, 0, false);

#else
                    GtkWidget *btn =
                        gtk_item_factory_get_widget(item_factory,
                        "/Options/Three Junctions");
                    if (btn) {
                        unsigned old = GTK_CHECK_MENU_ITEM(btn)->active;
                        GTK_CHECK_MENU_ITEM(btn)->active = (njj == 3 ? 1 : 0);
                        if (GTK_CHECK_MENU_ITEM(btn)->active != old) {
                            diag.diagram();
                            queue_redraw();
                        }
                    }
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_l1), la);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_l2), lb);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_l3), lc);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_l4), ld);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_frac), f);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_i1), ix);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_i2), i2);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_i3), iy);
                    gtk_spin_button_set_value(
                        GTK_SPIN_BUTTON(app.entry_coup), c);
                    // this triggers refresh
#endif
                }
            }
        }
        fclose(fp);
        return (false);
    }
    return (true);
}


static char *
get_data_file(const char *fname)
{
    char buf[256];
    char *home = getenv("HOME");
    if (home) {
        char *t = home + strlen(home) - 1;
        if (*t == '/'
#ifdef WIN32_NATIVE
                || *t == '\\'
#endif
                )
            sprintf(buf, "%s%s", home, fname);
        else
            sprintf(buf, "%s/%s", home, fname);
    }
    else
        strcpy(buf, *fname == '.' ? fname+1 : fname);
    return (strdup(buf));
}


static void
errmsg(const char *msg, int level)
{
#ifdef WIN32_NATIVE
    MessageBox(0, msg, level ? "Error" : "Warning",
        level ? MB_ICONSTOP : MB_ICONWARNING);
#else
    (void)level;
    fprintf(stderr, "%s\n", msg);
#endif
}

#ifdef WIN32_NATIVE

void
spinbtn::init(double val, double del, double minv, double maxv, int numd)
{
    value = val;
    delta = del;
    minval = minv;
    maxval = maxv;
    numdgt = numd;
    char buf[64];
    sprintf(buf, "%.*f", numdgt, value);
    Edit_SetText(entry, buf);
}


void
spinbtn::setval(bool down)
{
    double d = delta;
/*
    if (GRX->lastPressTime) {
        unsigned int dt = time(0) - GRX->lastPressTime;
        if (dt > 5)
            d *= 100;
        else if (dt > 2)
            d *= 10;
    }
    */

    if (down) {
        value -= d;
        if (value < minval)
            value = minval;
    }
    else {
        value += d;
        if (value > maxval)
            value = maxval;
    }
    char buf[64];
    sprintf(buf, "%.*f", numdgt, value);
    Edit_SetText(entry, buf);
}


void
spinbtn::setval(double val)
{
    if (val < minval)
        val = minval;
    else if (val > maxval)
        val = maxval;
    value = val;

    char buf[64];
    sprintf(buf, "%.*f", numdgt, value);
    Edit_SetText(entry, buf);
}

#endif
