
/**************************************************************************
 *
 * IPAT - Interferometer Threshold Curve Plotter
 *
 * Stephen R. Whiteley (stevew@srware.com)
 * 
 * Copyright (c) 2002 Whiteley Research Inc.  All Rights Reserved. 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************
 * $Id: ipat.h,v 2.6 2002/08/10 23:45:42 stevew Exp $
 ***************************************************************************/

#ifndef VERSION
#define VERSION "0.0"
#endif

/**************************************************************************

 THIS IS FREE SOFTWARE, NO GUARANTEES, NO LIABILITY, NO NOTHING!

 This program plots the interference patterns of two and three
 Josephson junction interferometers under the assumption that the
 junctions are perfect RSJ with no coupled flux.

 Copyright (C) Stephen R. Whiteley 1992
 Please send bug reports or comments to stevew@srware.com

 **************************************************************************/

#ifndef IPAT_H
#define IPAT_H

#include "screen.h"

#ifdef WIN32_NATIVE

struct spinbtn
{
    float minval;
    float maxval;
    float delta;
    float value;
    int numdgt;
    HWND entry;
    HWND ud;

    void init(double, double, double, double, int);
    void setval(bool);
    void setval(double);
};

#endif

#ifdef WIN32
int msw_ListPrinters(int*, char ***);
const char *msw_RawFileToPrinter(const char*, const char*);
#endif


// Container for the main application widgets
//
struct panel
{
    panel()
    {
        Scale = 2;
        Coupling = 1.0;
        flags.rel_x = -1;
        flags.rel_y = -1;
    }

    bool init();
    void init_recall_menu();
    void print_popup();
    void dialog_popup(unsigned int);
    void text_display_popup(char*);
    void extpts(FILE*);
    void delpts(void);
    bool save(const char*, bool);
    bool recall(const char*);
    void printscreen();

    int    Scale;            // scale factor, 1,2,4,8 where 8 is 1 period
    double Period;           // interference pattern period, 100% coupling
    double Coupling;         // current coupling ratio
    double Ic;               // maximum critical current
    double Order1,Order2;    // three jj phase offset parameters

    // printer interface
    int tofile;
    char *tofilename;
    char *cmdtext;
    int diag_x;
    int diag_y;

#ifdef WIN32_NATIVE
    HWND shell;
    spinbtn entries[9];
#else
    GtkWidget *shell;
    GtkWidget *menubar;
    GtkWidget *entry_l1;
    GtkWidget *entry_l2;
    GtkWidget *entry_l3;
    GtkWidget *entry_l4;
    GtkWidget *entry_frac;
    GtkWidget *entry_i1;
    GtkWidget *entry_i2;
    GtkWidget *entry_i3;
    GtkWidget *entry_coup;
    GtkItemFactory *item_factory;
#endif

    screen scr;     // drawing area for plot
    screen diag;    // drawing area for diagram

    // display mode flag storage
    //
    struct
    {
        char c;
        char f;
        char n;
        char r;
        char s;
        char x;
        char y;
        char z;
        char threej;
        double rel_x;
        double rel_y;
    } flags;

    // linked list of x-y file points to plot
    //
    extra *pts;
};

#endif

